@@ -59,13 +59,15 @@
  **/
 
 typedef struct {
-	SoupURI *uri;
+	SoupURI     *uri;
 	SoupAddress *addr;
 
 	GSList      *connections;      /* CONTAINS: SoupConnection */
 	guint        num_conns;
 
 	guint        num_messages;
+
+	gboolean     ssl_fallback;
 } SoupSessionHost;
 
 typedef struct {
@@ -624,6 +626,8 @@ soup_session_class_init (SoupSessionClass *session_class)
 	 * SOUP_SESSION_ACCEPT_LANGUAGE:
 	 *
 	 * Alias for the #SoupSession:accept-language property, qv.
+	 *
+	 * Since: 2.30
 	 **/
 	g_object_class_install_property (
 		object_class, PROP_ACCEPT_LANGUAGE,
@@ -649,6 +653,8 @@ soup_session_class_init (SoupSessionClass *session_class)
 	 * SOUP_SESSION_ACCEPT_LANGUAGE_AUTO:
 	 *
 	 * Alias for the #SoupSession:accept-language-auto property, qv.
+	 *
+	 * Since: 2.30
 	 **/
 	g_object_class_install_property (
 		object_class, PROP_ACCEPT_LANGUAGE_AUTO,
@@ -1266,6 +1272,9 @@ connection_disconnected (SoupConnection *conn, gpointer user_data)
 		g_hash_table_remove (priv->conns, conn);
 		host->connections = g_slist_remove (host->connections, conn);
 		host->num_conns--;
+
+		if (soup_connection_get_ssl_fallback (conn))
+			host->ssl_fallback = TRUE;
 	}
 
 	g_signal_handlers_disconnect_by_func (conn, connection_disconnected, session);
@@ -1386,6 +1395,7 @@ soup_session_get_connection (SoupSession *session,
 		SOUP_CONNECTION_ASYNC_CONTEXT, priv->async_context,
 		SOUP_CONNECTION_TIMEOUT, priv->io_timeout,
 		SOUP_CONNECTION_IDLE_TIMEOUT, priv->idle_timeout,
+		SOUP_CONNECTION_SSL_FALLBACK, host->ssl_fallback,
 		NULL);
 	g_signal_connect (conn, "disconnected",
 			  G_CALLBACK (connection_disconnected),
@@ -1949,8 +1959,8 @@ soup_session_remove_feature_by_type (SoupSession *session, GType feature_type)
  * you want to see all features, you can pass %G_TYPE_SESSION_FEATURE
  * for @feature_type.)
  *
- * Return value: (transfer container): a list of features. You must
- * free the list, but not its contents
+ * Return value: (transfer container) (element-type Soup.SessionFeature):
+ * a list of features. You must free the list, but not its contents
  *
  * Since: 2.26
  **/
