@@ -20,6 +20,8 @@ resolve_callback (SoupAddress *addr, guint status, gpointer data)
 	}
 	printf ("\n");
 
+	g_object_unref (addr);
+
 	nlookups--;
 	if (nlookups == 0)
 		g_main_loop_quit (loop);
@@ -41,8 +43,8 @@ main (int argc, char **argv)
 	if (argc < 2)
 		usage ();
 
-	g_type_init ();
 	g_thread_init (NULL);
+	g_type_init ();
 
 	for (i = 1; i < argc; i++) {
 		addr = soup_address_new (argv[i], 0);
@@ -57,7 +59,7 @@ main (int argc, char **argv)
 	}
 
 	loop = g_main_loop_new (NULL, TRUE);
-	g_main_run (loop);
+	g_main_loop_run (loop);
 	g_main_loop_unref (loop);
 
 	return 0;
